import andromeda.exception


HTTP_METHODS = ['get', 'post', 'put', 'delete', 'options', 'head', 'patch']


class Endpoint(object):
    '''
    Declares an endpoint to be instated by the server.
     
    Each of the methods below represents a handler that is automatically registered as a flask route.
    
    Whenever a route is invoked, a new Endpoint instance is created. Therefore, endpoint classes CAN be stateful, but
    take notice that it is discouraged.

    Each of the created instances has two local variables: request and session. They are being passed from the server
    itself, and therefore any deriving class MUST accept them in their constructor.
    '''

    # the url for which the methods will be registered.
    # supports uri types and regexes, much like a flask route.
    # uri parts are passed as keyword arguments to the handlers.
    url = '/'

    def __init__(self, request, context):
        self.request = request
        self.context = context

    def get(self, **uri_parts):
        self._method_not_allowed()

    def post(self, **uri_parts):
        self._method_not_allowed()

    def put(self, **uri_parts):
        self._method_not_allowed()

    def delete(self, **uri_parts):
        self._method_not_allowed()

    def options(self, **uri_parts):
        self._method_not_allowed()

    def head(self, **uri_parts):
        self._method_not_allowed()

    def patch(self, **uri_parts):
        self._method_not_allowed()

    def _method_not_allowed(self):
        raise andromeda.exception.HTTPException(405, 'method not allowed')



import collections
import flask

import andromeda.exception


class _RequestParser(object):
    '''
    RequestParser is an imperative interface to describe the parameters expected by an endpoint handler.
    
    It is very similar in form to argparse.ArgumentParser().
    
    This is an internal baseclass, and cannot be used outside.
    '''

    # the source string represents the part of the request used for supplying the arguments for a given class.
    # it is used by several exceptions in order to denote to the user where they went wrong when building their
    # request. meant to be overriden.
    source = 'request'

    def __init__(self):
        self._arguments = {}

    def add_argument(self, name, help='', type=object, required=False, default=None):
        '''
        Adds an argument declaration to the current parser.
        
        :param name: name of the argument 
        :param help: internal explanation as for the argument's nature
        :param type: the expected type of the value
        :param required: whether or not this argument MUST be passed in the request
        :param default: default value, in case this argument was not passed in the request
        '''

        if name in self._arguments:
            raise Exception('Argument {0} defined more than once.'.format(name))

        self._arguments[name] = {'help': help, 'type': type, 'required': required, 'default': default}
        self._validate_rule(name, self._arguments[name])

    def parse_args(self):
        '''
        Parses the request and extracts the declared arguments.
        
        :return: an object whose fields match the expected arguments 
        '''

        ParsedArguments = collections.namedtuple('ParsedArguments',
                                                 [argname.replace('-', '_') for argname in self._arguments.keys()])

        argument_values = []
        for arg_name, arg_rule in self._arguments.items():
            arg_val = self._get_argument_value(arg_name)

            if arg_val is None:
                arg_val = arg_rule['default']

            if arg_val is None and arg_rule['required']:
                raise andromeda.exception.HTTPException(400, 'missing field "{0}" in {1}: {2}'.format(arg_name,
                                                                                                      self.source,
                                                                                                      arg_rule['help']))

            if arg_val is not None and not isinstance(arg_val, arg_rule['type']):
                raise andromeda.exception.HTTPException(400, 'field "{0}": wrong type. expected: {1}'.format(arg_name,
                                                                                                             arg_rule['type'].__name__))

            argument_values.append(arg_val)

        return ParsedArguments(*argument_values)

    def _validate_rule(self, argument_name, rule):
        '''
        After compiling a rule supplied in add_argument, makes sure that it has no conflicts.
        
        Some parsers can override this method in order to add their own validations.
        
        :param argument_name: name of the argument matching the rule 
        :param rule: the rule, as built by add_argument
        '''

        if rule['required'] and rule['default'] is not None:
            raise Exception('Invalid argument {0}: cannot be "required" and have default value.'.format(argument_name))

    def _get_argument_value(self, name):
        '''
        When parsing arguments, this method provides an interface to get an argument's value from the request.
        
        Parsers implement this method according to the part of the request they refer to. 
        
        :param name: name of the parameter to fetch
        '''
        raise NotImplementedError('Internal class: Please use one of the parser subclasses instead.')


class BodyParser(_RequestParser):
    '''
    This parser is used for parsing a request's body.
    '''

    source = 'body'

    def _get_argument_value(self, name):
        try:
            return flask.request.get_json(force=True).get(name)
        except:
            return None


class _TypelessRequestParser(_RequestParser):
    '''
    A RequestParser that does not provide type validation.
    
    Used for parsers which handle all-string sources.
    '''

    def _validate_rule(self, argument_name, rule):
        super(_TypelessRequestParser, self)._validate_rule(argument_name, rule)

        if rule['type'] != object:
            raise Exception('Invalid argument {0}: {1} arguments cannot have types.'.format(argument_name, self.source))


class HeadersParser(_TypelessRequestParser):
    '''
    This parser is used for parsing a request's headers.
    '''

    source = 'headers'

    def _get_argument_value(self, name):
        try:
            return flask.request.headers.get(name)
        except:
            return None


class QuerystringParser(_TypelessRequestParser):
    '''
    This parser is used for parsing a request's query string args.
    '''

    source = 'querystring'

    def _get_argument_value(self, name):
        try:
            return flask.request.args.get(name.replace('_', '-'))
        except:
            return None



class HTTPException(Exception):
    '''
    An Exception class that denotes an HTTP exception, and is parsed by the server.
    '''

    def __init__(self, status, message):
        super(HTTPException, self).__init__('[{0}] {1}'.format(status, message))
        self._status = status
        self._message = message

    @property
    def status(self):
        return self._status

    @property
    def message(self):
        return self._message



from setuptools import setup, find_packages


setup(name='andromeda',
      version='0.1.0',
      description='declarative, class-driven framework written around flask',
      author='Roy Sommer',
      url='https://www.github.com/illberoy/andromeda',
      packages=find_packages(),
      include_package_data=True,
      py_modules=['andromeda'],
      install_requires=['flask==0.12.2'])



import functools
import flask

from .endpoint import HTTP_METHODS, Endpoint
from .parsers import BodyParser, HeadersParser, QuerystringParser
from .exception import HTTPException
from .context import Context


class Server(object):

    def __init__(self, name, *args, **kwargs):
        self._app = flask.Flask(name, *args, **kwargs)
        self._app.errorhandler(404)(functools.partial(self._error_handler, 404, 'not found'))
        self._endpoints_count = 0
        self._context = Context()
        self._middlewares = []

    def run(self, port, debug=False):
        self._app.run('0.0.0.0', port, debug)

    def use(self, cls):
        if isinstance(cls, list):
            for endpoint in cls:
                self.use(endpoint)

        else:
            for method in HTTP_METHODS:
                self._app.add_url_rule(cls.url,
                                       'endpoint.{0}'.format(self._endpoints_count),
                                       view_func=functools.partial(self._endpoint_handler, cls, method),
                                       methods=[method])

                self._endpoints_count += 1

    def add_context(self, attribute_name, context_value):
        setattr(self._context, attribute_name, context_value)

    def middleware(self, middleware):
        self._middlewares.append(middleware(self._app))

    def _endpoint_handler(self, endpoint_cls, method, **uri_params):
        request = flask.request

        try:
            # create instance in the request and server contexts
            endpoint_instance = endpoint_cls(request, context)

            # run endpoint handler
            response = getattr(endpoint_instance, method)(**uri_params)

            # return the outcome
            return self._render_response(response)

        except HTTPException as err:
            # return rendered exception
            return self._error_handler(err.status, err.message, err)

        except Exception as err:
            # return rendered exception with 500
            return self._error_handler(500, str(err), err)

    def _error_handler(self, status, message, err=None):
        response = flask.make_response((flask.jsonify({'status': status, 'message': message}), status))

        for middleware in self._middlewares:
            middleware(response)

        return response

    def _render_response(self, response):
        if isinstance(response, tuple):
            response = list(response)
            response[0] = flask.jsonify(response[0])
            response = tuple(response)
        else:
            response = flask.jsonify(response)

        response = flask.make_response(response)

        for middleware in self._middlewares:
            middleware(response)

        return response
